package de.hawhamburg.reachability.tools.feeder;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.log4j.Logger;
import org.joda.time.Minutes;

import com.ibm.icu.util.Calendar;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.reachability.StopableObject;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;
import de.hawhamburg.reachability.util.ParameterUtil;

/**
 * Main thread to transform defined sensor stories (*.sgs) into series of ActiveMQ sensor events.<br><br>
 * 
 * This class takes, when started, the supplied simulation file (in "*.sgs" format) and maps it to sensor events on the 
 * ActiveMQ which can easily be received by a corresponding sensor adapter. This tool allows the simulation of complex 
 * scenarios to check the reaction of the reachability agent system. It may also serve as training environment for an 
 * learning based reasoner.  
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class SensorDataFeeder extends Thread implements StopableObject {

    /**
     * The default prefix for simulated sensor result topics on the ActiveMQ. 
     */
    public static final String TOPIC_ROOT               = "RSDF_";

    /**
     * The default path to the directory of the simulation files.
     */
    public static final String TEST_PATH                = "C:/Users/Kantak/Desktop/szenarios/";

    /**
     * The default name of the file that contains the story which shall be simulated.
     */
    public static final String TEST_FILE                = "master_szenario_I.sgs";

    /**
     * The default simulation speed in simulation minutes per real-time minute.
     */
    public static final int FEEDER_SPEED                = 200;

    /**
     * The default waiting time before simulation starts in milliseconds.
     */
    public static final int FEEDER_WAIT                 = 5000;

    /**
     * The default instruction to send the actual simulation time each simulation minute as time sensor event.
     */
    public static final boolean FEEDER_HEARTBEAT        = false;

    /**
     * The default instruction to write informations to console/logfile.
     */
    public static final boolean FEEDER_SILENT           = false;

    /**
     * The actual prefix for simulated sensor result topics on the ActiveMQ.
     */
    private String topicRoot                            = null;

    /**
     * The actual path to the directory of the simulation files.
     */
    private String path                                 = null;

    /**
     * The actual name of the file that contains the story which shall be simulated.
     */
    private String file                                 = null;

    /**
     * The actual simulation speed in simulation minutes per real-time minute.
     */
    private int speed                                   = 1;

    /**
     * The actual delay time in milliseconds before the simulation shall start.
     */
    private int wait                                    = 0;

    /**
     * The actual instruction to send the actual simulation time each simulation minute as time sensor event.
     */
    private boolean heartbeat                           = false;

    /**
     * Determines if this service shall print out debug output or not.
     */
    private boolean silent                              = false;

    /**
     * The actual opened output topics on the ActiveMQ as {@link Map} of sensor name to {@link LPPublisher}. 
     */
    private Map<String, LPPublisher> outTopics          = new HashMap<String, LPPublisher>();

    /**
     * Determines if the method {@link #shutDown()} has been called.
     */
    private boolean shutDown                            = false;
    
    /**
     * The list of threads, which shall be notified at change and simulation end.
     */
    private List<SensorDataFeederNotifiable> notifiable = new ArrayList<SensorDataFeederNotifiable>();
    
    /**
     * The actual list of sent {@link SensorDataFeederItem}.
     */
    private SortedMap<SensorDataFeederItem, String> log = new TreeMap<SensorDataFeederItem, String>();
    
    /**
     * The actual simulation story as {@link Map} of time index ({@link Integer}) to {@link SensorDataFeederItem}.
     */
    private Map<Integer, SensorDataFeederItem> story;
        
    /**
     * Initializes a new {@link SensorDataFeeder} object with supplied command line arguments.<br><br>
     * 
     * <b>Defined command line arguments:</b><br>
     * <hr>
     * <table>
     *      <tr>
     *          <td><i>--storyfile</i></td>
     *          <td>Name of the used story file</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TEST_FILE})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--storypath</i></td>
     *          <td>Path to the used story file</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TEST_PATH})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederwait</i></td>
     *          <td>Delay time before start in milliseconds</td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederchannelroot</i></td>
     *          <td>Default AMQ topic prefix</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TOPIC_ROOT})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederspeed</i></td>
     *          <td>Speed of simulation in simulation minutes per real minute</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_SPEED})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederheartbeat</i></td>
     *          <td>Instruction to send the actual simulation time each simulation minute as time sensor event.</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_HEARTBEAT})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feedersilent</i></td>
     *          <td>Determines if this service shall print out debug output or not.</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_SILENT})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--wrapped</i></td>
     *          <td>Informs the systems it wasn't started as stand-alone service and forces it to be silent.</td>
     *      </tr>
     * </table>
     *                       
     * @param args              The command line arguments (see method description for usage).  
     */
    public SensorDataFeeder(final String[] args) {
        // Parameter utilization
        String[] parmKeys = {"--storyfile",
                             "--storypath",
                             "--feederwait",
                             "--feederchannelroot",
                             "--feederspeed",
                             "--feederheartbeat",
                             "--feedersilent",
                             "--wrapped"};
        ParameterUtil p = ParameterUtil.fromArgs(args, parmKeys);
        
        if (!p.isset("--wrapped")) {
            // Remove all existing loggers
            Logger.getRootLogger().removeAllAppenders();
        }
        
        // Select parameters
        String pathIn       = p.fetchString("--storypath",          TEST_PATH);
        String fileIn       = p.fetchString("--storyfile",          TEST_FILE);
        String channel      = p.fetchString("--feederchannelroot",  TOPIC_ROOT);
        int speedIn         = p.fetchInt("--feederspeed",           FEEDER_SPEED);
        int waitIn          = p.fetchInt("--feederwait",            FEEDER_WAIT);
        boolean heartbeatIn = p.fetchBool("--feederheartbeat",      FEEDER_HEARTBEAT);
        boolean silentIn    = p.fetchBool("--feedersilent",         FEEDER_SILENT);
        
        if (p.isset("--wrapped")) {
            silentIn = true;
        }

        // Save parameters
        this.path       = pathIn;
        this.file       = fileIn;
        this.story      = this.loadStory(pathIn + fileIn);
        this.topicRoot  = channel;
        this.speed      = speedIn;
        this.wait       = waitIn;
        this.heartbeat  = heartbeatIn;
        this.silent     = silentIn;
    }
    
    /**
     * Initializes a new {@link SensorDataFeeder} object.
     * 
     * @param basepath          The path to the simulation story file as {@link String}.    
     * @param topicRootIn       The prefix of ActiveMQ topics for simulated sensor event output as 
     *                          {@link String}.
     * @param speedIn           The in simulation minutes per real-time minute as <code>int</code>.
     * @param waitIn            The delay in milliseconds before the simulation starts as 
     *                          <code>int</code>.
     * @param sendHeartbeat     The instruction to send a frequently heartbeat including the
     *                          simulation time or not as <code>boolean</code>.
     * @param beQuite           Indicates if this simulation shall run in silent mode or not.
     */
    public SensorDataFeeder(final String basepath,
                            final String topicRootIn,
                            final int speedIn,
                            final int waitIn,
                            final boolean sendHeartbeat,
                            final boolean beQuite) {
        this.path       = basepath.substring(0, basepath.lastIndexOf("\\") + 1);
        this.file       = basepath.substring(basepath.lastIndexOf("\\") + 1);
        this.story      = this.loadStory(basepath);
        this.topicRoot  = topicRootIn;
        this.speed      = speedIn;
        this.wait       = waitIn;
        this.heartbeat  = sendHeartbeat;
        this.silent     = beQuite;
    }

    /**
     * Loads and returns a story from a defined story file by using {@link SensorDataFeederItem#fromFile(String)}.
     * 
     * @param storyFilePath     The path to the file that contains the story which shall be loaded.
     * @return                  The story as {@link Map} of simulation time minute ({@link Integer}) to 
     *                          {@link SensorDataFeederItem} object.
     */
    private Map<Integer, SensorDataFeederItem> loadStory(final String storyFilePath) {
        return SensorDataFeederItem.fromFile(storyFilePath);
    }
    
    /**
     * Adds a new thread which shall be notified if this simulation feeder stops.
     * 
     * @param object            The thread which shall be notified as {@link SensorDataFeederNotifiable}.
     */
    public final void addNotifiable(final SensorDataFeederNotifiable object) {
        if (!this.notifiable.contains(object)) {
            this.notifiable.add(object);
        }
    }
    
    /**
     * Unregisters a thread form the list of threats that shall be notified if the simulation feeder stops.
     * 
     * @param object            The thread which shall be unregistered of the list of notifiables. 
     */
    public final void unregisterNotifiable(final SensorDataFeederNotifiable object) {
        if (this.notifiable.contains(object)) {
            this.notifiable.remove(object);
        }
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#isActive()
     * 
     * @return                  <code>true</code> if {@link #shutDown()} wasn't called, otherwise <code>false</code>.
     */
    public final boolean isActive() {
        return !this.shutDown;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        this.shutDown = true;
    }

    /**
     * Starts the simulation.
     */
    public final void run() {
        OutputUtil.info("*** STARTING SIMULATION ***");
        
        if (wait > 0) {
            OutputUtil.info(
                    " --> Feeder will start in " + wait + "ms due to parameter settings <-- "
            );
            try {
                Thread.sleep(wait);
            } catch (InterruptedException e) {
                OutputUtil.error("Unable to wait... sorry :S");
            }
        }
        
        SortedSet<Integer> keys = new TreeSet<Integer>(story.keySet());
        
        SensorDataFeederItem lastItem = null;
        SensorDataFeederItem value = null;
        
        for (Integer key : keys) {
            if (this.shutDown) {
                break;
            }
            
            value = story.get(key);
            
            if (lastItem == null) {
                lastItem = value;
            } else {
                sendData(lastItem);
                simulateTime(lastItem.getTime(), value.getTime(), speed);
                lastItem = value;
            }
        }
        
        if (lastItem != null && !this.shutDown) {
            sendData(lastItem);
        }
        
        OutputUtil.info("*** SIMULATION ENDED ***");
        end();
    }
    
    /**
     * Gracefully end the thread by disconnecting all open {@link LPPublisher} and notifying all registered 
     * {@link SensorDataFeederNotifiable} objects.
     */
    private void end() {
        // Call shut down for good measure
        this.shutDown();
        
        // Disconnect all open AMQ connections
        for (LPPublisher p : outTopics.values()) {
            if (p != null) {
                p.disconnect();
            }
        }
        
        // Unregister from registered notify objects
        for (SensorDataFeederNotifiable notif : this.notifiable) {
            notif.unregisterSimulationFeeder(this);
        }
    }
    
    /**
     * Sends a new {@link SensorDataFeederItem} as JSON-object via the ActiveMQ.
     * 
     * @param item              The new sensor event as {@link SensorDataFeederItem}.
     */
    private void sendData(final SensorDataFeederItem item) {
        String topicSuffix = "data";
        String currTime = new SimpleDateFormat("HH:mm:ss").format(Calendar.getInstance().getTime());
        
        if (!silent) {
            OutputUtil.debug("[" + currTime + "] -> " + item + " -> " + this.topicRoot + topicSuffix);
        }
        
        ConnectionUtil.setMessageOnTopic(getTopic(topicSuffix), item.toJsonObject());
        log.put(item.clone(), topicRoot + topicSuffix);
    }
    
    /**
     * Creates and returns a new {@link LPPublisher} object which holds an open ActiveMQ topic connection named as 
     * "{@link #topicRoot}+[SENSOR_TYPE_NAME]".  
     * 
     * @param type              The name of the sensor type as {@link String}.  
     * @return                  An {@link LPPublisher} as open topic on the ActiveMQ to publish the simulated sensor 
     *                          event.
     */
    private LPPublisher getTopic(final String type) {
        if (!outTopics.containsKey(type)) {
            LPPublisher p = ConnectionUtil.getPublisher(topicRoot + type, this);
            outTopics.put(type, p);
        }
        return outTopics.get(type);
    }
    
    /**
     * Returns the path where the simulation file is located.
     * 
     * @return                  The path where the actual simulation file is located as {@link String}.
     */
    public final String getPath() {
        return this.path;
    }

    /**
     * Return the actual simulation files name.
     * 
     * @return                  The actual simulation files name as {@link String}.
     */
    public final String getFile() {
        return this.file;
    }

    /**
     * Returns the actual ActiveMQ topic prefix.
     * 
     * @return                  The actual ActiveMQ topic prefix as {@link String}.
     */
    public final String getTopicRoot() {
        return this.topicRoot;
    }

    /**
     * Returns the actual simulation speed in simulation minutes per real-time minute.
     * 
     * @return                  The actual simulation speed in simulation minutes per real-time minute.
     */
    public final int getSpeed() {
        return this.speed;
    }

    /**
     * Returns the whole actual simulation story as list of {@link SensorDataFeederItem}.
     * 
     * @return                  The whole actual simulation story as list of {@link SensorDataFeederItem}.
     */
    public final Map<Integer, SensorDataFeederItem> getStory() {
        return this.story;
    }

    /**
     * Returns the log of sent simulation events as list of {@link SensorDataFeederItem}.
     * 
     * @return                  The log of sent simulation events as list of {@link SensorDataFeederItem}.
     */
    public final SortedMap<SensorDataFeederItem, String> getLog() {
        SortedMap<SensorDataFeederItem, String> cpy = new TreeMap<SensorDataFeederItem, String>();
        cpy.putAll(this.log);
        return  cpy;
    }

    /**
     * Determines whether the simulation sends the actual simulation time as simulated time sensor each minute or not. 
     * 
     * @return                  <code>true</code> if this simulation sends its actual simulation time as simulated time
     *                          sensor, otherwise <code>false</code>.
     */
    public final boolean sendsHeartbeats() {
        return this.heartbeat;
    }

    /**
     * Determines whether the simulation runs in silent mode or not.
     * 
     * @return                  <code>true</code> if the simulation runs in silent mode, otherwise <code>false</code>.
     */
    public final boolean isSilent() {
        return this.silent;
    }
    
    /**
     * Let the thread sleep, until the next sensor event shall be published onto the ActiveMQ.
     * 
     * @param last              The simulation time in minutes where the last event took place.
     * @param next              The simulation time in minutes where the next event shall start.
     * @param speedIn           The current simulation time in simulation minutes per real-time minute.
     */
    private void simulateTime(final int last,
                              final int next,
                              final int speedIn) {
        // Get duration of a single minute in milliseconds.
        long millisPerMinute = Minutes.ONE.toStandardDuration().getMillis();
        
        // Calculate the difference time in simulation minutes.
        int differenceInMinutes = next - last;
        try {
            for (int i = 0; i < differenceInMinutes; i++) {
                if (this.shutDown) {
                    return;
                }
                if (heartbeat) {
                    String ctime = SensorDataFeederItem.getStringTime(last + i).replace(":", "");
                    sendData(new SensorDataFeederItem("time", ctime + "00", last + i));
                }
                Thread.sleep((long) (((double) 1 / (double) speedIn) * millisPerMinute));
            }
        } catch (InterruptedException e) {
            OutputUtil.error(e.getMessage());
        }
    }

    /**
     * Main method to start the sensor data feeder as stand-alone thread.<br><br>
     * 
     * <b>Defined command line arguments:</b><br>
     * <hr>
     * <table>
     *      <tr>
     *          <td><i>--storyfile</i></td>
     *          <td>Name of the used story file</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TEST_FILE})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--storypath</i></td>
     *          <td>Path to the used story file</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TEST_PATH})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederwait</i></td>
     *          <td>Delay time before start in milliseconds</td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederchannelroot</i></td>
     *          <td>Default AMQ topic prefix</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#TOPIC_ROOT})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederspeed</i></td>
     *          <td>Speed of simulation in simulation minutes per real minute</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_SPEED})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feederheartbeat</i></td>
     *          <td>Instruction to send the actual simulation time each simulation minute as time sensor event.</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_HEARTBEAT})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--feedersilent</i></td>
     *          <td>Determines if this service shall print out debug output or not.</td>
     *      </tr>
     *      <tr>
     *          <td></td>
     *          <td><i>(Defaults to {@link SensorDataFeeder#FEEDER_SILENT})</i></td>
     *      </tr>
     *      <tr>
     *          <td><i>--wrapped</i></td>
     *          <td>Informs the systems it wasn't started as stand-alone service and forces it to be silent.</td>
     *      </tr>
     * </table>
     *
     * @param args              The command line arguments (see method description for usage).
     */
    public static void main(final String[] args) {
        // Create and start a new SensorDataFeeder thread.
        SensorDataFeeder feeder = new SensorDataFeeder(args);
        feeder.start();
    }
}
